searchData={"content_type":"text/markdown","items":[{"doc":"A tool for building in-order GenStage topologies for EventStoreDB\n\nBeeline provides a Broadway-like experience for building GenStage\ntopologies for consuming streams of events from EventStoreDB in-order,\nusually one event at a time. Beeline aims to close over the supervision\nand basic configuration of the producer(s), as well as some of the\nrun-of-the-mill procedure done in the `c:GenStage.init/1` callback of\nthe consumer such as linking the producer process(es).","ref":"Beeline.html","title":"Beeline","type":"module"},{"doc":"Beeline creates a topology of GenStage, GenServer, and Supervisor processes.\nThis topology looks like the following:\n\n```text\nSupervisor\n├── HealthChecker*\n└── StageSupervisor\n    ├── Producer*\n    └── Consumer\n```\n\nLet's break these down from the bottom up:\n\n* \"Consumer\" - the GenStage consumer module which invokes\n  `Beeline.start_link/2`, handles events, and increments stream\n  positions.\n* \"Producer*\" - one or more GenStage producers which feed the consumer.\n  These producers are declared with the `:producers` key and may either\n  be `Kelvin.InOrderSubscription`, `Volley.InOrderSubscription`, or\n  `Beeline.DummyProducer` producer modules.\n* \"StageSupervisor\" - a supervisor for the GenStage pipeline. This supervisor\n  has a `:transient` restart strategy so that if the GenStage pipeline halts\n  on an event it cannot handle, the `StageSupervisor` supervision tree is\n  brought down but not the entire supervision tree. This behavior is\n  desirable so that the health-checker process can continue reading the\n  stream positions and so that an operator can perform any necessary\n  manual intervention on the crashed supervision tree (for example,\n  skipping the failure event).\n* \"HealthChecker*\" - a GenServer which periodically polls the stream positions\n  of a producer. There is one health checker process per producer.\n* \"Supervisor\" - a top-level supervisor. This supervisor has a `:permanent`\n  restart strategy.\n\nSee the `start_link/2` documentation for a full configuration reference and\nexamples.","ref":"Beeline.html#module-the-beeline-topology","title":"The Beeline Topology - Beeline","type":"module"},{"doc":"Wraps an event in a subscription event packet\n\nThis can be useful for building test events to pass through the dummy\nproducer.","ref":"Beeline.html#as_subscription_event/2","title":"Beeline.as_subscription_event/2","type":"function"},{"doc":"Decodes the body of a subscription event\n\nThis function performs JSON decoding if necessary and converts maps with\nstring keys into maps keyed by atoms. This This can potentially lead to\natom exhaustion, but the allowed atom count is quite high and usually this\nconcern is only theoretical.","ref":"Beeline.html#decode_event/1","title":"Beeline.decode_event/1","type":"function"},{"doc":"@impl GenStage\n    def handle_events([subscription_event], _from, state) do\n      event = Beeline.decode_event(subscription_event)\n      # ..","ref":"Beeline.html#decode_event/1-examples","title":"Examples - Beeline.decode_event/1","type":"function"},{"doc":"Determines which producer emitted the subscription event\n\nThis can be useful in order to save stream positions when a consumer is\nsubscribed to multiple producers. Should be used in tandem with\n`stream_position/1`.","ref":"Beeline.html#producer/1","title":"Beeline.producer/1","type":"function"},{"doc":"Restarts the supervision tree of GenStages for the given Beeline topology\n\nThis can be useful for manual intervention by a human operator in a remote\nconsole session, if the GenStage supervision tree crashes and exceeds the\nretry limits.","ref":"Beeline.html#restart_stages/1","title":"Beeline.restart_stages/1","type":"function"},{"doc":"iex> Beeline.restart_stages(MyEventHandler)\n    :ok","ref":"Beeline.html#restart_stages/1-examples","title":"Examples - Beeline.restart_stages/1","type":"function"},{"doc":"Starts a Beeline topology","ref":"Beeline.html#start_link/2","title":"Beeline.start_link/2","type":"function"},{"doc":"* `:name` - The GenServer name for the topology. The topology will build on this\n  name, using it as a prefix.\n\n* `:producers` (`t:keyword/0`) - A list of producers to which the consumer should subscribe. See the\n  \"producer options\" section below for the schema.\n\n* `:get_stream_position` - A function to invoke in order to get the stream position for a producer.\n  This function should be a 1-arity function (anonymous or capture) where\n  the name of the producer is passed as the argument. This option may also\n  be passed as an MFA tuple where the producer name will be prepended to\n  the argument list. If this option is not provided, a default will be\n  fetched with `Application.fetch_env!(:beeline, :get_stream_position)`.\n  This configuration can be used to set a blanket function for all\n  beelines to use.\n\n* `:auto_subscribe?` - A function to invoke to determine whether each producer should\n  subscribe to events as it starts up. The argument passed is the\n  GenServer name of the producer. If this option is not provided,\n  a default will be fetched with\n  `Application.fetch_env!(:beeline, :auto_subscribe?)`.\n\n* `:subscribe_after` - A period in msec after initialization when each producer should\n  query the `:auto_subscribe?` function. The default value is `{Enum, :random, [3000..5000]}`.\n\n* `:spawn_health_checkers?` - Controls whether the topology should spawn the HealthChecker children.\n  It can be useful to disable this in `Mix.env() in [:dev, :test]` as the\n  health checker provides little or no value in those environments and\n  can produce many log lines. If this option is left blank, it will be\n  gotten from the application environment defaulting to `true` with\n  `Application.get_env(:beeline, :spawn_health_checkers?, true)`. The default value is `nil`.\n\n* `:health_check_interval` - How long the health checker processes should wait between polling\n  the stream positions. Can either be a function (MFA or 0-arity function)\n  or a non-negative integer. The value is treated as milliseconds. The default value is `51000`.\n\n* `:health_check_drift` - A noise to add to the interval specified with `:health_check_interval`.\n  This can be useful to allow that not all producers poll their positions\n  at the same time, which can reduce strain on the stream position store\n  and the EventStoreDB. Can either be a function (MFA or 0-arity function)\n  or a non-negative integer. The value is treated as milliseconds. If a\n  function is provided, it is invoked every time the health checker\n  process attempts to schedule the next poll. The default value is `{Enum, :random, [0..10000]}`.\n\n* `:test_mode?` - Controls whether the topology should start up in test mode. In test\n  mode, any adapters set in producer specs are switched out with\n  the `:dummy` adapter. If this option is left blank, it will be\n  gotten from the application environment defaulting to `false` with\n  `Application.get_env(:beeline, :test_mode?, false)`.\n\n* `:context` (`t:term/0`) - A user-defined data structure which is used as the initial state of\n  the GenStage consumer process. The default value is `nil`.\n\n* `:acceptable_behind_by` (`t:non_neg_integer/0`) - If listener is behind stream but by this number of events it is considered ok.\n  If in next check it is still behind in this range but didn't process any event\n  since last check, listener is considered stale or \"behind\". The default value is `5`.\n\n\n\n#### Producer options\n\n* `:adapter` - Required. The adapter module to use for creating the producer. Use `:kelvin` for\n  EventStoreDB v3-5, `:volley` for EventStoreDB v20+, and `:dummy` for\n  test cases.\n\n* `:stream_name` (`t:String.t/0`) - Required. The name of the EventStoreDB stream to which this producer should\n  subscribe for events.\n\n* `:connection` (`t:atom/0`) - Required. The module to use as a connection to form the subscription. When the\n  `:adapter` option is `:kelvin`, this should be an Extreme client module.\n  When the adapter is `:volley`, it should be a `Spear.Client` module.\n\n* `:name` (`t:atom/0`) - The full GenServer name to use for this producer. When this option is\n  not provided, the name will be a formula of the name of the consumer\n  and the key in the keyword list of producers. The default value is `nil`.\n\n* `:max_demand` (`t:pos_integer/0`) - The maximum number of events the consumer is allowed to request from\n  this producer. This option can be configured to allow batch processing. The default value is `1`.\n\n* `:min_demand` (`t:non_neg_integer/0`) - The minimum number of events the consumer can request at a time from\n  this producer. The default value is `0`.","ref":"Beeline.html#start_link/2-options","title":"Options - Beeline.start_link/2","type":"function"},{"doc":"defmodule MyEventHandler do\n      use Beeline\n\n      def start_link(_opts) do\n        Beeline.start_link(MyEventHandler,\n          name: MyEventHandler,\n          producers: [\n            default: [\n              name: MyEventHandler.EventListener,\n              stream_name: \"$ce-BoundedContext.AggregateName\",\n              connection: MyEventHandler.EventStoreDBConnection,\n              adapter: :kelvin\n            ]\n          ]\n        )\n      end\n\n      # .. GenStage callbacks\n\n      @impl GenStage\n      def handle_events([subscription_event], _from, state) do\n        # .. handle the events one-by-one ..\n\n        {:noreply, [], state}\n      end\n    end","ref":"Beeline.html#start_link/2-examples","title":"Examples - Beeline.start_link/2","type":"function"},{"doc":"Determines the stream position of the subscription event\n\nThis function prefers link stream positions if available. This means that if\nthe subscription from which the event is emitted is reading a projected\nstream such as a category stream, the returned stream position will reflect\nthe position in the projected stream instead of the origin stream.","ref":"Beeline.html#stream_position/1","title":"Beeline.stream_position/1","type":"function"},{"doc":"@impl GenStage\n    def handle_events([subscription_event], _from, state) do\n      # consume the events\n\n      MyApp.Repo.transaction(fn ->\n        # save some state\n\n        producer = Beeline.producer(subscription_event)\n        stream_position = Beeline.stream_position(subscription_event)\n        MyApp.StreamPosition.persist(producer, stream_position)\n      end)\n    end","ref":"Beeline.html#stream_position/1-examples","title":"Examples - Beeline.stream_position/1","type":"function"},{"doc":"Gives a set of events to a topology's dummy producer\n\nThis function can be used to test running events through a topology.\nIf there are multiple producers, one is picked at random.","ref":"Beeline.html#test_events/2","title":"Beeline.test_events/2","type":"function"},{"doc":"A GenStage producer built for testing beeline topologies\n\nThis producer can be used either by setting the producer option `:adapter`\nto `:dummy` or by setting the `:test_mode?` option to `true`.\n\nThis producer emits events which arrive via `GenServer.cast/2`. Use\n`Beeline.test_events/2` to send events to a topology's dummy producer.","ref":"Beeline.DummyProducer.html","title":"Beeline.DummyProducer","type":"module"},{"doc":"A GenServer which periodically polls a producer's stream positions and\nprocess\n\nThis GenServer emits `:telemetry` measurements which serve as an interface\nfor exporting this health-check information to an external monitoring\nservice.\n\n`Beeline.HealthChecker.Logger` is included as a reasonable default consumer\nfor this telemetry. You may wish to export this telemetry to another system\nsuch as appsignal or grafana in order to create alerts when processors fall\nbehind.","ref":"Beeline.HealthChecker.html","title":"Beeline.HealthChecker","type":"module"},{"doc":"* `[:beeline, :health_check, :stop]` - dispatched by each HealthChecker\n  process after polling the producer's position and process information\n    * Measurement: `%{duration: native_time}` - the time taken to perform\n      the position and process checks\n    * Metadata, a map with the following keys:\n        * `:producer` (module) - the producer module being measured\n        * `:alive?` (boolean) - whether the producer process is alive\n        * `:stream_name` (string) - the EventStoreDB stream name from which\n          the producer reads\n        * `:hostname` (string) - the hostname of the machine on which\n          the health checker process is being run\n        * `:interval` (integer) - the milliseconds the health checker process\n          has waited (minus drift) since the last poll\n        * `:drift` (integer) - the milliseconds used for drifting the interval\n          for the last poll\n        * `:measurement_time` (UTC datetime) - the time when the poll started\n        * `:prior_position` (integer) - the `:current_position` from the last\n          poll\n        * `:current_position` (integer) - the current stream position of\n          the producer\n        * `:head_position` (integer) - the stream position of the head\n          (the latest event) of the EventStoreDB stream\n        * `:auto_subscribe` (boolean) - the value of the producer's `:auto_subscribe?`\n          flag at the time of polling","ref":"Beeline.HealthChecker.html#module-telemetry","title":"Telemetry - Beeline.HealthChecker","type":"module"},{"doc":"A Task to attach a logger exporter for health-checker telemetry\n\nAttaches a telemetry handler which writes out how far a producer is behind\nthe head of the stream to the logger.\n\nThis task can be started in a supervision tree such as an app's application\nsupervision tree:\n\n```elixir\ndef start(_type, _args) do\n  children = [\n    Beeline.HealthChecker.Logger,\n    MyApp.MyBeelineTopology\n  ]\n  opts = [strategy: :one_for_one, name: MyApp.Supervisor]\n  Supervisor.start_link(children, opts)\nend\n```\n\nThe log messages are 'debug' level in a format of the producer name\nconcatenated with \"is up-to-date.\" if the producer's current stream position\nmatches the latest available stream position (within `acceptable_behind_by`).\n\nWhen producer is falling behind and falling behind more then in a previous check,\nit is logged in warn level with \"is behind: n events.\" and \"is behind more: n events\"\nwith `n` being the number of events producer is behind head of stream.\n\nOnce producer is behind the stream less then in a previous check and when it catches up with it,\nit is logged in `info` level as `is behind but catching up: n events` and `is caught up.` respectively.\n\nThe log messages also include metadata fields\n`:event_listener` - the name of the producer - and `delta`: the number\nof events by which the producer is behind.","ref":"Beeline.HealthChecker.Logger.html","title":"Beeline.HealthChecker.Logger","type":"module"},{"doc":"Returns a specification to start this module under a supervisor.\n\n`arg` is passed as the argument to `Task.start_link/1` in the `:start` field\nof the spec.\n\nFor more information, see the `Supervisor` module,\nthe `Supervisor.child_spec/2` function and the `t:Supervisor.child_spec/0` type.","ref":"Beeline.HealthChecker.Logger.html#child_spec/1","title":"Beeline.HealthChecker.Logger.child_spec/1","type":"function"},{"doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a\nChangelog](https://keepachangelog.com/en/1.0.0/), and this project adheres to\n[Semantic Versioning](https://semver.org/spec/v2.0.0.html).\n\n<!--","ref":"changelog.html","title":"Changelog","type":"extras"},{"doc":"-->","ref":"changelog.html#unreleased","title":"UNRELEASED - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#1-1-4-2023-09-18","title":"1.1.4 - 2023-09-18 - Changelog","type":"extras"},{"doc":"- Log everything health check related with `debug` level if `auto_subscribe` is `false`.","ref":"changelog.html#changed","title":"Changed - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#1-1-3-2023-07-21","title":"1.1.3 - 2023-07-21 - Changelog","type":"extras"},{"doc":"- Added `previous_status` to metadata of telemetry event.","ref":"changelog.html#added","title":"Added - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#1-1-2-2023-07-14","title":"1.1.2 - 2023-07-14 - Changelog","type":"extras"},{"doc":"- More accurate health check monitoring and logging","ref":"changelog.html#changed","title":"Changed - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#1-1-1-2023-04-04","title":"1.1.1 - 2023-04-04 - Changelog","type":"extras"},{"doc":"- \"caught up\" messages logged at debug level","ref":"changelog.html#changed","title":"Changed - Changelog","type":"extras"},{"doc":"This change adds the ability to use :global and Registry names for the process\nand topology.","ref":"changelog.html#1-1-0-2022-10-12","title":"1.1.0 - 2022-10-12 - Changelog","type":"extras"},{"doc":"This change is for stability only and does not present any functional\nchanges from v0.3.1.","ref":"changelog.html#1-0-0-2022-05-11","title":"1.0.0 - 2022-05-11 - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#0-3-1-2022-05-05","title":"0.3.1 - 2022-05-05 - Changelog","type":"extras"},{"doc":"- Use `wrap_function` to properly evaluate `auto_subscribe` value","ref":"changelog.html#fixed","title":"Fixed - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#0-3-0-2022-04-14","title":"0.3.0 - 2022-04-14 - Changelog","type":"extras"},{"doc":"- Added `auto_subscribe` field to telemetry event","ref":"changelog.html#added","title":"Added - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#0-2-0-2021-07-14","title":"0.2.0 - 2021-07-14 - Changelog","type":"extras"},{"doc":"- The extreme adapter for fetching the latest event number has been patched\n  to treat the extreme connection as a process instead of a module\n  - this allows one to use connections which are not modules\n- Fixed whitespace in ArgumentError message for invalid configuration","ref":"changelog.html#fixed","title":"Fixed - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#0-1-0-2021-06-15","title":"0.1.0 - 2021-06-15 - Changelog","type":"extras"},{"doc":"- Initial implementation published to Hex.pm","ref":"changelog.html#added","title":"Added - Changelog","type":"extras"},{"doc":"","ref":"changelog.html#0-0-0-2021-06-09","title":"0.0.0 - 2021-06-09 - Changelog","type":"extras"},{"doc":"- This project was generated by Gaas","ref":"changelog.html#added","title":"Added - Changelog","type":"extras"}],"producer":{"name":"ex_doc","version":"0.36.1"},"proglang":"elixir"}